/*
 *  ImgView  A simple image viewer
 *
 *  Copyright (C) 2013 Mats Peterson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *  Please send any comments/bug reports to
 *  matsp888@yahoo.com  (Mats Peterson)
 */

#include <xsimple.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <FreeImage.h>
#include <pcre.h>
#include "imgview.h"
#include "reader/common.h"

VARS v;

static char *helpstr =
    "MOVEMENT\n"
    "\tArrow: Pan   +Shift: ×2   +Ctrl: ×4\n"
    "\tHome: Center   +Ctrl: Reset Zoom and Pixel Aspect Ratio\n"
    "\tL Btn: Pan\n"
    "\tR Btn: Center\n\n"
    "ZOOM\n"
    "\tPg Up: ×1.25   +Shift: ×1.5   +Ctrl: ×2\n"
    "\tPg Dn: ÷1.25   +Shift: ÷1.5   +Ctrl: ÷2\n"
    "\tz, Z: Toggle Lock Zoom / Auto Zoom\n\n"
    "PIXEL ASPECT RATIO\n"
    "\tF2: 320 × 200 / 640 × 400   +Shift: 640 × 200"
                 "   +Ctrl: 320 × 400\n"
    "\tF3: 320 × 256 / 640 × 512   +Shift: 640 × 256"
                 "   +Ctrl: 320 × 512\n"
    "\tF4: 352 × 240 / 720 × 480   +Shift: 352 × 480\n"
    "\tF5: 352 × 288 / 720 × 576   +Shift: 352 × 576\n"
    "\tF6: 640 × 350\n"
    "\tF7: 720 × 348\n"
    "\tF8: Custom\n"
    "\tF9: 1:1\n"
    "\ta, A: Toggle Lock Pixel Aspect Ratio / Auto Pixel Aspect"
                 " Ratio\n\n"
    "RESAMPLING FILTER\n"
    "\t1–6: Box / Bilinear / B-Spline / Bicubic / Catmull-Rom /"
                 " Lanczos\n"
    "\t0: None\n\n"
    "INFO\n"
    "\ti: Image Info\n"
    "\tc: Image Comments\n"
    "\tx: Exif Metadata\n"
    "\tm: XMP Metadata\n"
    "\tM: All Metadata\n\n"
    "FILE\n"
    "\tn, N: Next/Last File in Current Directory\n"
    "\tp, P: Previous/First File in Current Directory\n"
    "\t+, -: Next/Previous Page in Multi-Page TIFF File\n"
    "\ts, S: Toggle Slideshow of Current Dir. with 5/10 Seconds"
                 " Interval\n\n"
    "OTHER\n"
    "\tf: Toggle Full Screen\n"
    "\tu: Program Status\n"
    "\tF1: Command Help\n"
    "\tq, Esc: Quit";

static char *fmtstr[] = {
    "XBM (X Bitmap)",
    "PBM (Portable Bitmap)",
    "PGM (Portable Graymap)",
    "PPM (Portable Pixmap)",
    "GIF (Graphics Interchange Format)",
    "JPEG",
    "TIFF",
    "PNG (Portable Network Graphics)",
    "PCX (ZSoft PC Paintbrush)"
};

static char *jpgfmtstr[] = {
    "JPEG JIF (JPEG Interchange Format)",
    "JPEG JFIF (JPEG File Interchange Format)",
    "JPEG Exif (Exchangeable Image File Format)"
};

static char *tiffmtstr[] = {
    "TIFF (Tagged Image File Format)",
    "TIFF Exif (Exchangeable Image File Format)"
};

static double parl[] = {
    1.,              /* 1:1 */
    12. / 10.,       /* 320x200 / 640x400 */
    24. / 10.,       /* 640x200 */             12. / 20.,     /* 320x400 */
    15. / 16.,       /* 320x256 / 640x512 */
    15. / 8.,        /* 640x256 */             15. / 32.,     /* 320x512 */
    4739. / 4320.,   /* 352x240 / 720x480 */   4739. / 8640., /* 352x480 */
    117. / 128.,     /* 352x288 / 720x576 */   117. / 256.,   /* 352x576 */
    13715. / 10000., /* 640x350 */
    15517. / 10000., /* 720x348 */
    1.,              /* Custom */
    1.               /* From image */
};

static char tspace[] = " ";


static void cleanup(void)
{
    if (*v.owd) chdir(v.owd);
}

static int extmatch(char *fname)
{
    static pcre *re = NULL;
    const char *err = NULL;
    int erroffs;
    int ovec[30];

    if (! re) re = pcre_compile("\\.(?:xbm|p[bgp]m|gif|jpe?g|jpe|"
            "jf?if|tiff?|png|pc[cx])$", PCRE_CASELESS, &err, &erroffs,
            NULL);
    if (pcre_exec(re, NULL, fname, strlen(fname), 0, 0, ovec, 30) > 0)
        return 1;
    else
        return 0;
}

static char* fmtsize(int size)
{
    static char buf[128];
    char tmp[50], *p1, *p2;
    int cnt = 0;

    sprintf(tmp, "%d", size);
    buf[sizeof(buf) - 1] = '\0';
    p1 = &tmp[strlen(tmp) - 1];
    p2 = &buf[sizeof(buf) - 2];
    while (p1 >= tmp) {
        if (cnt == 3) {
            char *p = tspace + sizeof(tspace) - 2;
            while (p >= tspace) { *p2-- = *p--; }
            cnt = 0;
        }
        *p2-- = *p1--;
        cnt++;
    }
    return ++p2;
}

static void errmsg(char *s)
{
    fprintf(stderr, "%s\n", s);
}

static void sighandler(int sig)
{
    v.timeout++;
}

static void mycls(void)
{
    if (! v.curimg) {
    	settitle(WINNAME);
        XFillRectangle(xv.d, xv.w, xv.gc, 0, 0, v.win_w, v.win_h);
        XSync(xv.d, False);
        return;
    }

    if (v.ulx > 0 && v.ulx > v.oulx)
        XFillRectangle(xv.d, xv.w, xv.gc, v.oulx, 0, v.ulx - v.oulx,
                v.win_h);
    if (v.uly > 0 && v.uly > v.ouly)
        XFillRectangle(xv.d, xv.w, xv.gc, 0, v.ouly, v.win_w,
                v.uly - v.ouly);
    if (v.lrx < (v.win_w - 1) && v.lrx < v.olrx)
        XFillRectangle(xv.d, xv.w, xv.gc, v.lrx + 1, 0, v.olrx - v.lrx,
                v.win_h);
    if (v.lry < (v.win_h - 1) && v.lry < v.olry)
	XFillRectangle(xv.d, xv.w, xv.gc, 0, v.lry + 1, v.win_w,
                v.olry - v.lry);
}

static void prgstatus(void)
{
    char infostr[1024];
    char *onoff_str[2] = {"Off", "On"};
    char *rfilt_str[8] = {"Box", "Bicubic", "Bilinear", "B-Spline",
            "Catmull-Rom", "Lanczos", "None"};
    sprintf(infostr,
            "Zoom: %g\n"
            "Lock Zoom: %s\n"
            "Auto Zoom: %s\n"
            "Pixel Aspect Ratio (h/w): %g\n"
            "Lock Pixel Aspect Ratio: %s\n"
            "Auto Pixel Aspect Ratio: %s\n"
            "Resampling Filter: %s",
            v.zoom, onoff_str[v.lockzoom], onoff_str[v.autozoom],
            parl[v.parn], onoff_str[v.lockpar], onoff_str[v.autopar],
            rfilt_str[v.rfilter]);
    message("Program Status", infostr, NULL);
}

static void imginfo(void)
{
    if (v.curimg) {
        char timestr[80], pagestr[80] = "";
        struct tm *tm;
        tm = localtime(&v.statbuf.st_mtime);
        strftime(timestr, sizeof(timestr), "%Y‐%m‐%d %H:%M:%S", tm);
        if (v.npages > 1) {
            sprintf(pagestr, "Page: %d of %d\n", v.page + 1, v.npages);
        }
        sprintf(v.infostr,
                "File Name: %s\n"
                "File Size: %s bytes\n"
                "File Last Modified: %s\n"
                "Image File Format: %s\n"
                "%s"
                "Dimensions: %d × %d pixels",
                v.files[v.si[v.fnum]], fmtsize(v.statbuf.st_size), timestr,
                (v.fmt == FMT_JPEG) ? jpgfmtstr[v.jpgfmt] :
                (v.fmt == FMT_TIFF) ? tiffmtstr[v.tiffmt] : fmtstr[v.fmt],
                pagestr, v.img->width, v.img->height);
        if (*v.xinfostr) {
            strcat(v.infostr, "\n"); strcat(v.infostr, v.xinfostr);
        }
    } else {
        sprintf(v.infostr, "File Name: %s", v.files[v.si[v.fnum]]);
    }
    message("Image Info", v.infostr, NULL);
}

static void imgmeta(int mode)
{
    char cmd[1024];
    char efile[1024];
    char *p1, *p2;
    int i, len;

    p1 = v.files[v.si[v.fnum]];
    p2 = efile;
    len = strlen(p1);
    for (i = 0; i < len; i++) {
        if (*p1 == '"' || *p1 == '`' || *p1 == '\\' || *p1 == '$')
            *p2++ = '\\';
        *p2++ = *p1++;
    }
    *p2 = '\0';

    sprintf(cmd, "ivmeta \"%s\" %s", efile, mode == 2 ? "-ExifTool:All" :
            mode == 1 ? "XMP:All" : "EXIF:All");

    system(cmd);
}

static void putimg(int docls)
{
    v.ulx = v.img_x - v.curimg->width/2;
    v.uly = v.img_y - v.curimg->height/2;
    v.lrx = v.ulx + v.curimg->width - 1;
    v.lry = v.uly + v.curimg->height - 1;

    if (docls) mycls();

    if (! xv.pmap) {
        GC putgc;
        xv.pmap = mkpixmap(v.curimg->width, v.curimg->height,
                v.curimg->depth);
        if (v.curimg->depth == 1) {
            static GC mgc = NULL;
            if (! mgc) mgc = XCreateGC(xv.d, xv.pmap, 0, NULL);
            if (v.black == 1) {
                XSetBackground(xv.d, mgc, 0);
                XSetForeground(xv.d, mgc, 1);
            } else {
                XSetBackground(xv.d, mgc, 1);
                XSetForeground(xv.d, mgc, 0);
            }
            putgc = mgc;
	} else
            putgc = xv.gc;
        XPutImage(xv.d, xv.pmap, putgc, v.curimg, 0, 0, 0, 0,
                v.curimg->width, v.curimg->height);
    }

    if (v.curimg->depth == 1)
        XCopyPlane(xv.d, xv.pmap, xv.w, v.monogc, 0, 0,
                v.curimg->width, v.curimg->height, v.ulx, v.uly, 1);
    else
        XCopyArea(xv.d, xv.pmap, xv.w, xv.gc, 0, 0, v.curimg->width,
                v.curimg->height, v.ulx, v.uly);

    v.oulx = v.ulx; v.ouly = v.uly; v.olrx = v.lrx; v.olry = v.lry;
}

static void doautopar(void)
{
    if (v.par != 1.) {
        parl[14] = v.par;
        v.parn = 14;
    } else {
        int w = v.img->width; int h = v.img->height;
        v.parn = ((w == 320 && h == 200) || (w == 640 && h == 400)) ? 1 :
              (w == 640 && h == 200) ? 2 :
              (w == 320 && h == 400) ? 3 :
              ((w == 320 && h == 256) || (w == 640 && h == 512)) ? 4 :
              (w == 640 && h == 256) ? 5 :
              (w == 320 && h == 512) ? 6 :
              ((w == 352 && h == 240) || (w == 720 && h == 480)) ? 7 :
              (w == 352 && h == 480) ? 8 :
              ((w == 352 && h == 288) || (w == 720 && h == 576)) ? 9 :
              (w == 352 && h == 576) ? 10 :
              (w == 640 && h == 350) ? 11 :
              (w == 720 && h == 348) ? 12 :
              0;
    }
}

static int setpar(void)
{
    char ps[80];
    char *nps;
    sprintf(ps, "%g", parl[13]);
    nps = input("Set Pixel Aspect Ratio", "Pixel Aspect Ratio (h/w)", ps);
    if (nps) {
        parl[13] = atof(nps);
        return 1;
    } else
        return 0;
}

static void doautozoom(void)
{
    double h2 = (double)(v.img->height) * parl[v.parn];
    v.zoom = (double)v.win_w / (double)(v.img->width);
    if ((int)(v.zoom * h2) > v.win_h) v.zoom = (double)v.win_h / h2;
}

static void resample_freeimage(void)
{
    FIBITMAP *fdib, *frdib;
    int fbpl;
    int pad;
    int x, y;

    fdib = FreeImage_Allocate(v.img->width, v.img->height,
            v.img->depth == 24 ? 32 : v.img->depth,
            FI_RGBA_RED_MASK, FI_RGBA_GREEN_MASK, FI_RGBA_BLUE_MASK);

    if (v.img->depth == 1) {
        RGBQUAD pal[2];
        if (v.black == 1) {
            pal[0].rgbRed = pal[0].rgbGreen = pal[0].rgbBlue = 255;
            pal[1].rgbRed = pal[1].rgbGreen = pal[1].rgbBlue = 0;
        } else {
            pal[0].rgbRed = pal[0].rgbGreen = pal[0].rgbBlue = 0;
            pal[1].rgbRed = pal[1].rgbGreen = pal[1].rgbBlue = 255;
        }
        pal[0].rgbReserved = pal[1].rgbReserved = 0;
        memcpy(FreeImage_GetPalette(fdib), pal, sizeof(RGBQUAD) * 2);
    }

    fbpl = FreeImage_GetLine(fdib);
    if (fbpl % 4) fbpl += (4 - (fbpl % 4));

    for (y = 0; y < v.img->height; y++) {
        memcpy(FreeImage_GetBits(fdib) + fbpl * y,
                v.img->data + v.img->bytes_per_line * y,
                v.img->bytes_per_line);
    }

    frdib = FreeImage_Rescale(fdib, v.zoom_w, v.zoom_h, v.rfilter);
    FreeImage_Unload(fdib);

    v.zimg = mkximage(v.zoom_w, v.zoom_h, 24, &pad);

    if (v.img->depth == 1) {
        fbpl = FreeImage_GetLine(frdib);
        if (fbpl % 4) fbpl += (4 - (fbpl % 4));
        for (y = 0; y < v.zimg->height; y++) {
            RGBQUAD *pal;
            unsigned char *bp;
            pal = FreeImage_GetPalette(frdib);
            bp = FreeImage_GetBits(frdib) + fbpl * y;
            dp = (unsigned char *)v.zimg->data +
                    v.zimg->bytes_per_line * y;
            idp = (unsigned int *)dp;
            for (x = 0; x < v.zimg->width; x++) {
                DOPIXEL(pal[*bp].rgbRed, pal[*bp].rgbGreen,
                       pal[*bp].rgbBlue, 255);
                bp++;
            }
        }
    } else {
        memcpy(v.zimg->data, FreeImage_GetBits(frdib),
                v.zimg->bytes_per_line * v.zimg->height);
    }

    FreeImage_Unload(frdib);
}

static void zoomimg(int docls)
{
    if (v.zimg) { rmximage(v.zimg); v.zimg = NULL; }
    if (xv.pmap) { rmpixmap(xv.pmap); xv.pmap = 0; }

    v.zoom_w = (int)((double)(v.img->width) * v.zoom + .5);
    v.zoom_h = (int)((double)(v.img->height) * v.zoom * parl[v.parn] + .5);

    if (v.zoom_w != v.img->width || v.zoom_h != v.img->height) {
        if (v.rfilter != FILTER_NONE)
            resample_freeimage();
        else {
            int pad;
            double xscale, yscale;

            v.zimg = mkximage(v.zoom_w, v.zoom_h, v.img->depth, &pad);

            xscale = (double)(v.img->width) / (double)(v.zimg->width);
            yscale = (double)(v.img->height) / (double)(v.zimg->height);

            if (v.img->depth == 1) {
                int x, y;
                memset(v.zimg->data, 0, v.zimg->bytes_per_line *
                        v.zimg->height);
                for (y = 0; y < v.zimg->height; y++) {
                    unsigned char *p1, *p2, dmask;
                    p1 = (unsigned char *)v.img->data +
                            (int)(y * yscale) * v.img->bytes_per_line;
                    p2 = (unsigned char *)v.zimg->data +
                            y * v.zimg->bytes_per_line;
                    for (x = 0, dmask = 0x80; x < v.zimg->width;
                            x++, dmask >>= 1) {
                        int sx;
                        if (! dmask) { p2++; dmask = 0x80; }
                        sx = x * xscale;
                        if (*(p1 + sx / 8) & (0x80 >> (sx % 8)))
                            *p2 |=  dmask;
                    }
                }
            } else {
                int x, y;
                unsigned int *p1 = (unsigned int *)v.img->data;
                unsigned int *p2 = (unsigned int *)v.zimg->data;
                unsigned int *tp;
                for (y = 0; y < v.zimg->height; y++) {
                    tp = p1 + (int)(y * yscale) * v.img->width;
                    for (x = 0; x < v.zimg->width; x++)
                        *p2++ = *(tp + (int)(x * xscale));
                }
            }
        }
        v.curimg = v.zimg;
    } else
        v.curimg = v.img;

    putimg(docls);
}

static void getfmt(int fd)
{
    unsigned char magic[4];
    unsigned short *magicsp;
    unsigned int *magicip;

    if (read(fd, magic, 4) < 4) { v.fmt = -1; return; }
    magicsp = (unsigned short *)magic;
    magicip = (unsigned int *)magic;
    lseek(fd, 0, SEEK_SET);

    if (magic[0] == 'P' && (magic[1] == '1' || magic[1] == '4'))
        v.fmt = FMT_PBM;
    else if (magic[0] == 'P' && (magic[1] == '2' || magic[1] == '5'))
        v.fmt = FMT_PGM;
    else if (magic[0] == 'P' && (magic[1] == '3' || magic[1] == '6'))
        v.fmt = FMT_PPM;
    else if (*magicip == 0x38464947)
        v.fmt = FMT_GIF;
    else if (*magicsp == 0xd8ff)
        v.fmt = FMT_JPEG;
    else if (*magicip == 0x2a004d4d || *magicip == 0x002a4949)
        v.fmt = FMT_TIFF;
    else if (*magicip == 0x474e5089)
        v.fmt = FMT_PNG;
    else if (magic[0] == 0x0a)
        v.fmt = FMT_PCX;
    else
        v.fmt = FMT_XBM;
}

static void readfile(char *fname)
{
    char *tail = NULL, tmp[512];
    int fd;
    int i;

    if (v.img) { rmximage(v.img); v.img = NULL; }
    if (v.zimg) { rmximage(v.zimg); v.zimg = NULL; }
    if (xv.pmap) { rmpixmap(xv.pmap); xv.pmap = 0; }
    v.curimg = NULL;

    v.ippi[0] = v.ippi[1] = 72.;
    v.par = 1.;
    *v.xinfostr = '\0';
    if (fname) v.page = v.npages = 0;

    for (i = 0; i < v.ncomments; i++) {
        free(v.comments[i].key); free(v.comments[i].text);
    }
    v.ncomments = 0;

    v.black = 1; v.white = 0;

    v.iprof = NULL;
    v.iproflen = 0;

    mycls();

    if (! fname) fname = v.files[v.si[v.fnum]];
    if ((tail = strrchr(fname, '/'))) tail++; else tail = fname; 

    if ((fd = open(fname, O_RDONLY)) < 0) {
        perror("open");
        goto err;
    }
    if (fstat(fd, &v.statbuf) < 0) {
        perror("fstat");
        goto err;
    }

    getfmt(fd);
    switch (v.fmt) {
        case FMT_XBM:
            close(fd); v.img = readxbm(fname); break;
        case FMT_PBM:
        case FMT_PGM:
        case FMT_PPM:
            v.img = readpnm(fd); break;
        case FMT_GIF:
            v.img = readgif(fd); break;
        case FMT_JPEG:
            v.img = readjpeg(fd, &v.jpgfmt); break;
        case FMT_TIFF:
            v.img = readtiff(fd, fname, v.page, &v.npages, &v.tiffmt);
            break;
        case FMT_PNG:
            v.img = readpng(fd); break;
        case FMT_PCX:
            v.img = readpcx(fd); break;
        default:
            break;
    }

    if (! v.img) goto err;

    if (v.ncomments) {
        int i, len = 0;
        for (i = 0; i < v.ncomments; i++) {
            if (i) len += 1;
            if (*(v.comments[i].key)) {
                char *p;
                len += strlen(v.comments[i].key) + 2;
                for (p = v.comments[i].text; *p; p++) {
                    len++; if (*p == '\n') len++;
                }
            } else
                len += strlen(v.comments[i].text);
        }
        if (v.commentstr) free(v.commentstr);
        v.commentstr = (char *)malloc(len + 1);
        *v.commentstr = '\0';
        for (i = 0; i < v.ncomments; i++) {
            if (i) strcat(v.commentstr, "\n");
            if (*(v.comments[i].key)) {
                char *p1, *p2;
                strcat(v.commentstr, v.comments[i].key);
                strcat(v.commentstr, ": ");
                p1 = v.commentstr + strlen(v.commentstr);
                for (p2 = v.comments[i].text; *p2; p2++) {
                    *p1++ = *p2;
                    if (*p2 == '\n') *p1++ = '\t';
                }
                *p1 = '\0';
            } else
                strcat(v.commentstr, v.comments[i].text);
        }
    }

    v.curimg = v.img;
    v.img_x = v.win_w/2; v.img_y = v.win_h/2;

    sprintf(tmp, "%s – %s [%d × %d]", WINNAME, tail, v.img->width,
            v.img->height);
    settitle(tmp);

    if (!v.ippi[0] || !v.ippi[1]) v.ippi[0] = v.ippi[1] = 72.;

    if (v.autopar) doautopar(); else if (! v.lockpar) v.parn = 0;
    if (v.autozoom) doautozoom(); else if (! v.lockzoom) v.zoom = 1.;
    if (v.zoom != 1. || v.parn != 0) zoomimg(0);
            else putimg(0);
    return;

err:
    sprintf(tmp, "%s – %s [ERROR]", WINNAME, tail);
    settitle(tmp);
}

static int compare(const void *a, const void *b)
{
    int i;
    i = strcoll(v.files[*(unsigned short *)a],
            v.files[*(unsigned short *)b]);
    return i;
}

static void _readdir(void)
{
    DIR *dir;
    struct dirent *de;
    int i;

    for (i = 0; i < v.nfiles; i++) free(v.files[i]);
    v.nfiles = 0;

    if ((dir = opendir(v.cwd))) {
        while ((de = readdir(dir)) && v.nfiles < MAX_FILES) {
            if (extmatch(de->d_name)) {
                v.files[v.nfiles] = strdup(de->d_name);
                v.si[v.nfiles] = (unsigned short)v.nfiles;
                v.nfiles++;
            }
        }
        closedir(dir);
    }

    if (v.nfiles) qsort(v.si, v.nfiles, sizeof(unsigned short), compare);
}

static void usage(void)
{
    fprintf(stderr, "usage: imgview file\n");
    exit(1);
}

static void init(int argc, char **argv)
{
    char rcname[512], *p;
    FILE *f;

    atexit(cleanup);

    if (argc == 1) usage();

    setlocale(LC_ALL, "");

    v.nfiles = 0;
    v.fnum = 0;
    v.win_x = WIN_X; v.win_y = WIN_Y; v.win_w = WIN_W; v.win_h = WIN_H;
    strcpy(v.bgcolor_s, BGCOLOR);
    strcpy(v.tbgcolor_s, TBGCOLOR);
    strcpy(v.bwbgcolor_s, BWBGCOLOR);
    strcpy(v.bwfgcolor_s, BWFGCOLOR);
    strcpy(v.rgbprofile, RGBPROFILE);
    strcpy(v.grayprofile, GRAYPROFILE);
    strcpy(v.cmykprofile, CMYKPROFILE);
    strcpy(v.outprofile, OUTPROFILE);
    v.xymove = XYMOVE;
    v.ncomments = 0;
    v.commentstr = NULL;
    v.npages = 0;
    v.zoom = v.par = 1.;
    v.lockzoom = v.autozoom = v.lockpar = v.autopar = v.parn =
            v.slide = v.timeout = 0;
    v.rfilter = FILTER_NONE;

    signal(SIGALRM, sighandler);

    sprintf(rcname, "%s/.imgviewrc", getenv("HOME"));
    if ((f = fopen(rcname, "r")))
        readconfig(f);

    if ((p = getenv("IMGVIEW_BGCOLOR")))
        strcpy(v.bgcolor_s, p);
    if ((p = getenv("IMGVIEW_TBGCOLOR")))
        strcpy(v.tbgcolor_s, p);
    if ((p = getenv("IMGVIEW_BWBGCOLOR")))
        strcpy(v.bwbgcolor_s, p);
    if ((p = getenv("IMGVIEW_BWFGCOLOR")))
        strcpy(v.bwfgcolor_s, p);
    if ((p = getenv("IMGVIEW_RGBPROFILE")))
        strcpy(v.rgbprofile, p);
    if ((p = getenv("IMGVIEW_GRAYPROFILE")))
        strcpy(v.grayprofile, p);
    if ((p = getenv("IMGVIEW_CMYKPROFILE")))
        strcpy(v.cmykprofile, p);
    if ((p = getenv("IMGVIEW_OUTPROFILE")))
        strcpy(v.outprofile, p);

    if (strlen(v.outprofile)) {
        if (! strcmp(v.outprofile, "BUILTIN_SRGB_PROFILE"))
            v.hOutProfile = cmsCreate_sRGBProfile();
        else
            v.hOutProfile = cmsOpenProfileFromFile(v.outprofile, "r");
        if (! v.hOutProfile) {
            fprintf(stderr,
                    "Error: Couldn't open output ICC profile `%s'\n",
                    v.outprofile);
            exit(1);
        }
        v.oprof = proftomem(v.hOutProfile, &v.oproflen);
        if (strlen(v.rgbprofile)) {
            if (! (v.hRGBProfile =
                    cmsOpenProfileFromFile(v.rgbprofile, "r"))) {
                fprintf(stderr,
                        "Error: Couldn't open RGB ICC profile `%s'\n",
                        v.rgbprofile);
                exit(1);
            }
        }
        if (strlen(v.grayprofile)) {
            if (! (v.hGrayProfile =
                    cmsOpenProfileFromFile(v.grayprofile, "r"))) {
                fprintf(stderr,
                        "Error: Couldn't open Gray ICC profile `%s'\n",
                        v.grayprofile);
                exit(1);
            }
        }
        if (strlen(v.cmykprofile)) {
            if (! (v.hCMYKProfile =
                    cmsOpenProfileFromFile(v.cmykprofile, "r"))) {
                fprintf(stderr,
                        "Error: Couldn't open CMYK ICC profile `%s'\n",
                        v.cmykprofile);
                exit(1);
            }
        }
    }

    xinit(APPNAME, CLASSNAME, WINNAME, v.win_x, v.win_y, v.win_w, v.win_h, 1,
            v.bgcolor_s, v.bgcolor_s, ButtonPressMask);

    v.win_w = xv.width; v.win_h = xv.height;
    v.bgcolor = getnamedcolor(v.bgcolor_s, &v.bgcolor_rgb);
    v.tbgcolor = getnamedcolor(v.tbgcolor_s, &v.tbgcolor_rgb);
    v.bwbgcolor = getnamedcolor(v.bwbgcolor_s, NULL);
    v.bwfgcolor = getnamedcolor(v.bwfgcolor_s, NULL);

    XSetForeground(xv.d, xv.gc, v.bgcolor);

    v.monogc = XCreateGC(xv.d, xv.w, 0, NULL);
    XSetBackground(xv.d, v.monogc, v.bwbgcolor);
    XSetForeground(xv.d, v.monogc, v.bwfgcolor);
}

int main(int argc, char **argv)
{
    char fname[FILENAME_MAX];
    struct timeval t;
    char s[FILENAME_MAX + 30], *p;
    int i;

    init(argc, argv);

    strcpy(fname, argv[1]);
    if (! extmatch(fname)) {
        sprintf(s, "%s: Unknown file name extension", fname);
        errmsg(s);
        return 1;
    }

    getcwd(v.cwd, FILENAME_MAX); strcpy(v.owd, v.cwd);
    if ((p = strrchr(fname, '/'))) {
        char dir[FILENAME_MAX];
        strcpy(dir, fname);
        *(dir + (p - fname) + 1) = '\0';
        if (chdir(dir) < 0) {
            perror("chdir");
            return 1;
        }
        getcwd(v.cwd, FILENAME_MAX);
        strcpy(fname, p + 1);
    }

    _readdir();

    v.fnum = -1;
    for (i = 0; i < v.nfiles; i++)
        if (! strcmp(v.files[v.si[i]], fname)) { v.fnum = i; break; }
    if (v.fnum == -1) {
        char s[FILENAME_MAX + 30];
        sprintf(s, "%s: File not found", fname);
        errmsg(s);
        return 1;
    }

    readfile(v.files[v.si[v.fnum]]);

    for (;;) {
        XEvent e;
        static int config_notify = 0;

        if (v.slide && v.timeout) {
            if (++v.fnum == v.nfiles) v.fnum = 0;
            readfile(v.files[v.si[v.fnum]]);
            v.timeout = 0; alarm(v.delay);
        }

	while (getevent(&e)) {
            XKeyEvent *ke;
            XButtonEvent *be;
            int k = 0, s;

            switch (e.type) {
                case ConfigureNotify:
                    config_notify++;
                    break;
		case Expose:
                    if (v.curimg) {
                        for (;;) {
                            XExposeEvent *ee = (XExposeEvent *)&e;
                            if (v.curimg->depth == 1)
                                XCopyPlane(xv.d, xv.pmap, xv.w, v.monogc,
                                        ee->x - v.ulx, ee->y - v.uly,
                                        ee->width, ee->height, ee->x,
                                        ee->y, 1);
                            else
                                XCopyArea(xv.d, xv.pmap, xv.w, xv.gc,
                                        ee->x - v.ulx, ee->y - v.uly,
                                        ee->width, ee->height, ee->x,
                                        ee->y);
                            if (! ee->count)
                                break;
                            XNextEvent(xv.d, &e);
                        }
                    }
		    break;
                case KeyPress:
                    ke = (XKeyEvent *)&e;
                    k = XLookupKeysym(ke, 0);
                    s = ke->state;
                    switch (k) {
                        case XK_Left:
                            if (! v.curimg) break;
                            v.img_x += v.xymove * (s & ControlMask ? 4 :
                                    s & ShiftMask ? 2 : 1);
                            putimg(1);
                            break;
                        case XK_Right:
                            if (! v.curimg) break;
                            v.img_x -= v.xymove * (s & ControlMask ? 4 :
                                    s & ShiftMask ? 2 : 1);
                            putimg(1);
                            break;
                        case XK_Up:
                            if (! v.curimg) break;
                            v.img_y += v.xymove * (s & ControlMask ? 4 :
                                    s & ShiftMask ? 2 : 1);
                            putimg(1);
                            break;
                        case XK_Down:
                            if (! v.curimg) break;
                            v.img_y -= v.xymove * (s & ControlMask ? 4 :
                                    s & ShiftMask ? 2 : 1);
                            putimg(1);
                            break;
                        case XK_Prior:
                            if (! v.curimg) break;
                            v.zoom *= (s & ControlMask ? 2. :
                                    s & ShiftMask ? 1.5 : 1.25);
                            zoomimg(1);
                            break;
                        case XK_Next:
                            if (! v.curimg) break;
                            v.zoom /= (s & ControlMask ? 2. :
                                    s & ShiftMask ? 1.5 : 1.25);
                            zoomimg(1);
                            break;
                        case XK_Home:
                            if (! v.curimg) break;
                            v.img_x = v.win_w/2;
                            v.img_y = v.win_h/2;
                            if (s & ControlMask) {
                                v.zoom = 1.0; zoomimg(1);
                            } else {
                                putimg(1);
                            }
                            break;
                        case XK_F1:
                            message("Command Help", helpstr, NULL);
                            break;
                        case XK_F2:
                            if (! v.curimg) break;
                            if (s & ControlMask)
                                v.parn = 3;
                            else if (s & ShiftMask)
                                v.parn = 2;
                            else
                                v.parn = 1;
                            if (v.autozoom) doautozoom();
                            zoomimg(1);
                            break;
                        case XK_F3:
                            if (! v.curimg) break;
                            if (s & ControlMask)
                                v.parn = 6;
                            else if (s & ShiftMask)
                                v.parn = 5;
                            else
                                v.parn = 4;
                            if (v.autozoom) doautozoom();
                            zoomimg(1);
                            break;
                        case XK_F4:
                            if (! v.curimg) break;
                            if (s & ShiftMask)
                                v.parn = 8;
                            else
                                v.parn = 7;
                            if (v.autozoom) doautozoom();
                            zoomimg(1);
                            break;
                        case XK_F5:
                            if (! v.curimg) break;
                            if (s & ShiftMask)
                                v.parn = 10;
                            else
                                v.parn = 9;
                            if (v.autozoom) doautozoom();
                            zoomimg(1);
                            break;
                        case XK_F6:
                            if (! v.curimg) break;
                            v.parn = 11;
                            if (v.autozoom) doautozoom();
                            zoomimg(1);
                            break;
                        case XK_F7:
                            if (! v.curimg) break;
                            v.parn = 12;
                            if (v.autozoom) doautozoom();
                            zoomimg(1);
                            break;
                        case XK_F8:
                            if (! v.curimg) break;
                            if (setpar()) {
                                v.parn = 13;
                                if (v.autozoom) doautozoom();
                                zoomimg(1);
                            }
                            break;
                        case XK_F9:
                            if (! v.curimg) break;
                            v.parn = 0;
                            if (v.autozoom) doautozoom();
                            zoomimg(1);
                            break;
                        case XK_p:
                            if (v.nfiles > 1) {
                                int ofnum = v.fnum;
                                v.fnum = s & ShiftMask ? 0 :
                                        v.fnum > 0 ? v.fnum - 1 : v.fnum;
                                if (v.fnum != ofnum) {
                                    readfile(v.files[v.si[v.fnum]]);
                                    if (v.slide) gettimeofday(&t, NULL);
                                }
                            }
                            break;
                        case XK_n:
                            if (v.nfiles > 1) {
                                int ofnum = v.fnum;
                                v.fnum = s & ShiftMask ? v.nfiles - 1 :
                                        v.fnum < v.nfiles - 1 ?
                                        v.fnum + 1 : v.fnum;
                                if (v.fnum != ofnum)  { 
                                    readfile(v.files[v.si[v.fnum]]);
                                    if (v.slide) gettimeofday(&t, NULL);
                                }
                            }
                            break;
                        case XK_equal:
                            if (v.fmt == FMT_TIFF &&
                                    v.page < v.npages - 1) {
                                v.page++; readfile(NULL);
                            }
                            break;
                        case XK_minus:
                            if (v.fmt == FMT_TIFF && v.page > 0) {
                                v.page--; readfile(NULL);
                            }
                            break;
                        case XK_z:
                            if (! v.curimg) break;
                            if (s & ShiftMask) {
                                v.autozoom ^= 1;
                                if (v.autozoom) {
                                    v.lockzoom = 0; doautozoom();
                                } else
                                    v.zoom = 1.;
                                zoomimg(1);
                            } else
                                v.lockzoom ^= 1;
                            break;
                        case XK_a:
                            if (! v.curimg) break;
                            if (s & ShiftMask) {
                                v.autopar ^= 1;
                                if (v.autopar) {
                                    v.lockpar = 0; doautopar();
                                } else
                                    v.parn = 0;
                                if (v.autozoom) doautozoom();
                                zoomimg(1);
                            } else
                                v.lockpar ^= 1;
                            break;
                        case XK_1:
                            if (! v.curimg) break;
                            v.rfilter = FILTER_BOX;
                            zoomimg(1);
                            break;
                        case XK_2:
                            if (! v.curimg) break;
                            v.rfilter = FILTER_BILINEAR;
                            zoomimg(1);
                            break;
                        case XK_3:
                            if (! v.curimg) break;
                            v.rfilter = FILTER_BSPLINE;
                            zoomimg(1);
                            break;
                        case XK_4:
                            if (! v.curimg) break;
                            v.rfilter = FILTER_BICUBIC;
                            zoomimg(1);
                            break;
                        case XK_5:
                            if (! v.curimg) break;
                            v.rfilter = FILTER_CATMULLROM;
                            zoomimg(1);
                            break;
                        case XK_6:
                            if (! v.curimg) break;
                            v.rfilter = FILTER_LANCZOS3;
                            zoomimg(1);
                            break;
                        case L'0':
                            if (! v.curimg) break;
                            v.rfilter = FILTER_NONE;
                            zoomimg(1);
                            break;
                        case XK_i:
                            if (v.curimg) imginfo();
                            break;
                        case XK_c:
                            if (v.curimg && v.ncomments)
                                message("Image Comments", v.commentstr,
                                NULL);
                            break;
                        case XK_x:
                            if (! v.curimg) break;
                            if (((v.fmt == FMT_JPEG &&
                                    v.jpgfmt == FMT_JPEG_EXIF) ||
                                (v.fmt == FMT_TIFF &&
                                    v.tiffmt == FMT_TIFF_EXIF)))
                                imgmeta(0);
                            break;
                        case XK_m:
                            if (! v.curimg) break;
                            if (s & ShiftMask)
                                imgmeta(2);
                            else
                                imgmeta(1);
                            break;
                        case XK_s:
                            if (! v.nfiles) break;
                            v.slide ^= 1;
                            v.delay = s & ShiftMask ? 10 : 5;
                            break;
                        case XK_f:
                            if (v.curimg)
                                toggle_fullscreen();
                            break;
                        case XK_u:
                            prgstatus();
                            break;
                        case XK_q:
                        case XK_Escape:
                            exit(0);
                     }
                     break;
                 case ButtonPress:
                     be = (XButtonEvent *)&e;
                     switch (be->button) {
                         case 1:
                             if (! v.curimg) break;
                             v.img_x += (v.win_w / 2) - be->x;
                             v.img_y += (v.win_h / 2) - be->y;
                             putimg(1);
                             break;
                         case 2:
                             break;
                         case 3:
                             if (! v.curimg) break;
                             v.img_x = v.win_w / 2;
                             v.img_y = v.win_h / 2;
                             putimg(1);
                             break;
                     }
            }
            if (v.slide && ((e.type == KeyPress && k != XK_Shift_L &&
                    k != XK_Shift_R && k != XK_Control_L &&
                    k != XK_Control_R) || e.type == ButtonPress)) {
                v.timeout = 0; alarm(v.delay);
            }
        }

        if (config_notify) {
            v.win_x = xv.x; v.win_y = xv.y;
            if (xv.width != v.win_w || xv.height != v.win_h) {
                v.win_w = xv.width; v.win_h = xv.height;
                v.img_x = v.win_w/2; v.img_y = v.win_h/2;
                if (v.curimg) {
                    if (v.autozoom) {
                        doautozoom(); zoomimg(1);
                    } else
                        putimg(1);
                }
            }
            config_notify = 0;
        }

        usleep(10000);
    }
}
